#!/usr/bin/env python
# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _hmc module of the zhmcclient_mock package.
"""

from __future__ import absolute_import, print_function

import unittest
import re

from zhmcclient_mock._hmc import FakedHmc, \
    FakedActivationProfileManager, FakedActivationProfile, \
    FakedAdapterManager, FakedAdapter, \
    FakedCpcManager, FakedCpc, \
    FakedHbaManager, FakedHba, \
    FakedLparManager, FakedLpar, \
    FakedNicManager, FakedNic, \
    FakedPartitionManager, FakedPartition, \
    FakedPortManager, FakedPort, \
    FakedVirtualFunctionManager, FakedVirtualFunction, \
    FakedVirtualSwitchManager, FakedVirtualSwitch, \
    FakedBaseManager, FakedBaseResource


class FakedHmcTests(unittest.TestCase):
    """All tests for the zhmcclient_mock.FakedHmc class."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')

    def test_repr(self):

        # The test approach is to check the repr() result for each attribute
        # that is shown in the result, but leaving some flexibility in how that
        # is formatted.

        # Bring everything into one line, because regexp is line-oriented.
        act_repr = repr(self.hmc).replace('\n', '\\n')

        self.assertRegexpMatches(
            act_repr,
            r'.*FakedHmc\s+at\s+0x{id:08x}\s+\(\\n.*'.
            format(id=id(self.hmc)))

        self.assertRegexpMatches(
            act_repr,
            r'.*\shmc_name\s*=\s*{hmc_name!r}\\n.*'.
            format(hmc_name=self.hmc.hmc_name))

        self.assertRegexpMatches(
            act_repr,
            r'.*\shmc_version\s*=\s*{hmc_version!r}\\n.*'.
            format(hmc_version=self.hmc.hmc_version))

        self.assertRegexpMatches(
            act_repr,
            r'.*\sapi_version\s*=\s*{api_version!r}\\n.*'.
            format(api_version=self.hmc.api_version))

        self.assertRegexpMatches(
            act_repr,
            r'.*\scpcs\s*=\s*FakedCpcManager\s.*')
        # TODO: Check content of `cpcs`

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_resources\s*=\s.*')
        # TODO: Check content of `_resources`

    def test_hmc(self):
        self.assertEqual(self.hmc.hmc_name, 'fake-hmc')
        self.assertEqual(self.hmc.hmc_version, '2.13.1')
        self.assertEqual(self.hmc.api_version, '1.8')
        self.assertIsInstance(self.hmc.cpcs, FakedCpcManager)

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()

        self.assertEqual(len(cpcs), 0)

    def test_hmc_1_cpc(self):
        cpc1_in_props = {'name': 'cpc1'}

        # the function to be tested:
        cpc1 = self.hmc.cpcs.add(cpc1_in_props)

        cpc1_out_props = cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()

        self.assertEqual(len(cpcs), 1)
        self.assertEqual(cpcs[0], cpc1)

        self.assertIsInstance(cpc1, FakedCpc)
        self.assertEqual(cpc1.properties, cpc1_out_props)
        self.assertEqual(cpc1.manager, self.hmc.cpcs)

    def test_hmc_2_cpcs(self):
        cpc1_in_props = {'name': 'cpc1'}

        # the function to be tested:
        cpc1 = self.hmc.cpcs.add(cpc1_in_props)

        cpc1_out_props = cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })

        cpc2_in_props = {'name': 'cpc2'}

        # the function to be tested:
        cpc2 = self.hmc.cpcs.add(cpc2_in_props)

        cpc2_out_props = cpc2_in_props.copy()
        cpc2_out_props.update({
            'object-id': cpc2.oid,
            'object-uri': cpc2.uri,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()

        self.assertEqual(len(cpcs), 2)
        # We expect the order of addition to be maintained:
        self.assertEqual(cpcs[0], cpc1)
        self.assertEqual(cpcs[1], cpc2)

        self.assertIsInstance(cpc1, FakedCpc)
        self.assertEqual(cpc1.properties, cpc1_out_props)
        self.assertEqual(cpc1.manager, self.hmc.cpcs)

        self.assertIsInstance(cpc2, FakedCpc)
        self.assertEqual(cpc2.properties, cpc2_out_props)
        self.assertEqual(cpc2.manager, self.hmc.cpcs)

    def test_res_dict(self):
        cpc1_in_props = {'name': 'cpc1'}
        adapter1_in_props = {'name': 'osa1', 'adapter-family': 'hipersockets'}
        port1_in_props = {'name': 'osa1_1'}

        rd = {
            'cpcs': [
                {
                    'properties': cpc1_in_props,
                    'adapters': [
                        {
                            'properties': adapter1_in_props,
                            'ports': [
                                {'properties': port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }

        # the function to be tested:
        self.hmc.add_resources(rd)

        cpcs = self.hmc.cpcs.list()

        self.assertEqual(len(cpcs), 1)

        cpc1 = cpcs[0]
        cpc1_out_props = cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })
        self.assertIsInstance(cpc1, FakedCpc)
        self.assertEqual(cpc1.properties, cpc1_out_props)
        self.assertEqual(cpc1.manager, self.hmc.cpcs)

        cpc1_adapters = cpc1.adapters.list()

        self.assertEqual(len(cpc1_adapters), 1)
        adapter1 = cpc1_adapters[0]

        adapter1_ports = adapter1.ports.list()

        self.assertEqual(len(adapter1_ports), 1)
        port1 = adapter1_ports[0]

        adapter1_out_props = adapter1_in_props.copy()
        adapter1_out_props.update({
            'object-id': adapter1.oid,
            'object-uri': adapter1.uri,
            'status': 'active',
            'network-port-uris': [port1.uri],
        })
        self.assertIsInstance(adapter1, FakedAdapter)
        self.assertEqual(adapter1.properties, adapter1_out_props)
        self.assertEqual(adapter1.manager, cpc1.adapters)

        port1_out_props = port1_in_props.copy()
        port1_out_props.update({
            'element-id': port1.oid,
            'element-uri': port1.uri,
        })
        self.assertIsInstance(port1, FakedPort)
        self.assertEqual(port1.properties, port1_out_props)
        self.assertEqual(port1.manager, adapter1.ports)


class FakedBaseTests(unittest.TestCase):
    """All tests for the FakedBaseManager and FakedBaseResource classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')

        self.cpc1_oid = '42-abc-543'
        self.cpc1_uri = '/api/cpcs/%s' % self.cpc1_oid

        self.cpc1_in_props = {
            # All properties that are otherwise defaulted (but with non-default
            # values), plus 'name'.
            'object-id': self.cpc1_oid,
            'object-uri': self.cpc1_uri,
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'status': 'service',
            'name': 'cpc1',
        }
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                },
            ]
        }
        self.hmc.add_resources(rd)
        self.cpc_manager = self.hmc.cpcs
        self.cpc_resource = self.hmc.cpcs.list()[0]
        self.cpc1_out_props = self.cpc1_in_props.copy()

    def test_manager_attr(self):
        """Test FakedBaseManager attributes."""

        self.assertIsInstance(self.cpc_manager, FakedBaseManager)

        self.assertEqual(self.cpc_manager.hmc, self.hmc)
        self.assertEqual(self.cpc_manager.parent, self.hmc)
        self.assertEqual(self.cpc_manager.resource_class, FakedCpc)
        self.assertEqual(self.cpc_manager.base_uri, '/api/cpcs')
        self.assertEqual(self.cpc_manager.oid_prop, 'object-id')
        self.assertEqual(self.cpc_manager.uri_prop, 'object-uri')

    def test_resource_attr(self):
        """Test FakedBaseResource attributes."""

        self.assertIsInstance(self.cpc_resource, FakedBaseResource)

        self.assertEqual(self.cpc_resource.manager, self.cpc_manager)
        self.assertEqual(self.cpc_resource.properties, self.cpc1_out_props)
        self.assertEqual(self.cpc_resource.oid,
                         self.cpc1_out_props['object-id'])
        self.assertEqual(self.cpc_resource.uri,
                         self.cpc1_out_props['object-uri'])

    def test_manager_repr(self):
        """Test FakedBaseManager.__repr__()."""

        # The test approach is to check the repr() result for each attribute
        # that is shown in the result, but leaving some flexibility in how that
        # is formatted.

        # Bring everything into one line, because regexp is line-oriented.
        act_repr = repr(self.cpc_manager).replace('\n', '\\n')

        self.assertRegexpMatches(
            act_repr,
            r'.*{classname}\s+at\s+0x{id:08x}\s+\(\\n.*'.
            format(classname=self.cpc_manager.__class__.__name__,
                   id=id(self.cpc_manager)))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_hmc\s*=\s*{hmc_classname}\s+at\s+0x{hmc_id:08x}\\n.*'.
            format(hmc_classname=self.cpc_manager.hmc.__class__.__name__,
                   hmc_id=id(self.cpc_manager.hmc)))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_parent\s*=\s*{p_classname}\s+at\s+0x{p_id:08x}\\n.*'.
            format(p_classname=self.cpc_manager.parent.__class__.__name__,
                   p_id=id(self.cpc_manager.parent)))

        m = re.match(r'.*\s_resource_class\s*=\s*([^\\]+)\\n.*', act_repr)
        if not m:
            raise AssertionError("'_resource_class = ...' did not match "
                                 "in: {!r}".format(act_repr))
        act_resource_class = m.group(1)
        self.assertEqual(act_resource_class,
                         repr(self.cpc_manager.resource_class))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_base_uri\s*=\s*{_base_uri!r}\\n.*'.
            format(_base_uri=self.cpc_manager.base_uri))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_oid_prop\s*=\s*{_oid_prop!r}\\n.*'.
            format(_oid_prop=self.cpc_manager.oid_prop))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_uri_prop\s*=\s*{_uri_prop!r}\\n.*'.
            format(_uri_prop=self.cpc_manager.uri_prop))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_resources\s*=\s.*')
        # TODO: Check content of `_resources`

    def test_resource_repr(self):
        """Test FakedBaseResource.__repr__()."""

        # The test approach is to check the repr() result for each attribute
        # that is shown in the result, but leaving some flexibility in how that
        # is formatted.

        # Bring everything into one line, because regexp is line-oriented.
        act_repr = repr(self.cpc_resource).replace('\n', '\\n')

        self.assertRegexpMatches(
            act_repr,
            r'.*{classname}\s+at\s+0x{id:08x}\s+\(\\n.*'.
            format(classname=self.cpc_resource.__class__.__name__,
                   id=id(self.cpc_resource)))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_manager\s*=\s*{m_classname}\s+at\s+0x{m_id:08x}\\n.*'.
            format(m_classname=self.cpc_resource.manager.__class__.__name__,
                   m_id=id(self.cpc_resource.manager)))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_manager._parent._uri\s*=\s*{p_uri!r}\\n.*'.
            format(p_uri=self.cpc_resource.manager.parent.uri))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_uri\s*=\s*{_uri!r}\\n.*'.
            format(_uri=self.cpc_resource.uri))

        self.assertRegexpMatches(
            act_repr,
            r'.*\s_properties\s*=\s.*')
        # TODO: Check content of `_properties`


class FakedActivationProfileTests(unittest.TestCase):
    """All tests for the FakedActivationProfileManager and
    FakedActivationProfile classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.resetprofile1_in_props = {'name': 'resetprofile1'}
        self.imageprofile1_in_props = {'name': 'imageprofile1'}
        self.loadprofile1_in_props = {'name': 'loadprofile1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'reset_activation_profiles': [
                        {'properties': self.resetprofile1_in_props},
                    ],
                    'image_activation_profiles': [
                        {'properties': self.imageprofile1_in_props},
                    ],
                    'load_activation_profiles': [
                        {'properties': self.loadprofile1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedActivationProfileManager:
        self.hmc.add_resources(rd)

    def test_profiles_attr(self):
        """Test CPC '*_activation_profiles' attributes."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # Test reset activation profiles

        self.assertIsInstance(cpc1.reset_activation_profiles,
                              FakedActivationProfileManager)
        self.assertEqual(cpc1.reset_activation_profiles.profile_type, 'reset')
        self.assertRegexpMatches(cpc1.reset_activation_profiles.base_uri,
                                 r'/api/cpcs/[^/]+/reset-activation-profiles')

        # Test image activation profiles

        self.assertIsInstance(cpc1.image_activation_profiles,
                              FakedActivationProfileManager)
        self.assertEqual(cpc1.image_activation_profiles.profile_type, 'image')
        self.assertRegexpMatches(cpc1.image_activation_profiles.base_uri,
                                 r'/api/cpcs/[^/]+/image-activation-profiles')

        # Test load activation profiles

        self.assertIsInstance(cpc1.load_activation_profiles,
                              FakedActivationProfileManager)
        self.assertEqual(cpc1.load_activation_profiles.profile_type, 'load')
        self.assertRegexpMatches(cpc1.load_activation_profiles.base_uri,
                                 r'/api/cpcs/[^/]+/load-activation-profiles')

    def test_profiles_list(self):
        """Test list() of FakedActivationProfileManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # Test reset activation profiles

        resetprofiles = cpc1.reset_activation_profiles.list()

        self.assertEqual(len(resetprofiles), 1)
        resetprofile1 = resetprofiles[0]
        resetprofile1_out_props = self.resetprofile1_in_props.copy()
        resetprofile1_out_props.update({
            'element-id': resetprofile1.oid,
            'element-uri': resetprofile1.uri,
        })
        self.assertIsInstance(resetprofile1, FakedActivationProfile)
        self.assertEqual(resetprofile1.properties, resetprofile1_out_props)
        self.assertEqual(resetprofile1.manager, cpc1.reset_activation_profiles)

        # Test image activation profiles

        imageprofiles = cpc1.image_activation_profiles.list()

        self.assertEqual(len(imageprofiles), 1)
        imageprofile1 = imageprofiles[0]
        imageprofile1_out_props = self.imageprofile1_in_props.copy()
        imageprofile1_out_props.update({
            'element-id': imageprofile1.oid,
            'element-uri': imageprofile1.uri,
        })
        self.assertIsInstance(imageprofile1, FakedActivationProfile)
        self.assertEqual(imageprofile1.properties, imageprofile1_out_props)
        self.assertEqual(imageprofile1.manager, cpc1.image_activation_profiles)

        # Test load activation profiles

        loadprofiles = cpc1.load_activation_profiles.list()

        self.assertEqual(len(loadprofiles), 1)
        loadprofile1 = loadprofiles[0]
        loadprofile1_out_props = self.loadprofile1_in_props.copy()
        loadprofile1_out_props.update({
            'element-id': loadprofile1.oid,
            'element-uri': loadprofile1.uri,
        })
        self.assertIsInstance(loadprofile1, FakedActivationProfile)
        self.assertEqual(loadprofile1.properties, loadprofile1_out_props)
        self.assertEqual(loadprofile1.manager, cpc1.load_activation_profiles)

    def test_profiles_add(self):
        """Test add() of FakedActivationProfileManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        resetprofiles = cpc1.reset_activation_profiles.list()
        self.assertEqual(len(resetprofiles), 1)

        resetprofile2_in_props = {'name': 'resetprofile2'}

        # the function to be tested:
        new_resetprofile = cpc1.reset_activation_profiles.add(
            resetprofile2_in_props)

        resetprofiles = cpc1.reset_activation_profiles.list()
        self.assertEqual(len(resetprofiles), 2)

        resetprofile2 = [p for p in resetprofiles
                         if p.properties['name'] ==
                         resetprofile2_in_props['name']][0]

        self.assertEqual(new_resetprofile.properties, resetprofile2.properties)
        self.assertEqual(new_resetprofile.manager, resetprofile2.manager)

        resetprofile2_out_props = resetprofile2_in_props.copy()
        resetprofile2_out_props.update({
            'element-id': resetprofile2.oid,
            'element-uri': resetprofile2.uri,
        })
        self.assertIsInstance(resetprofile2, FakedActivationProfile)
        self.assertEqual(resetprofile2.properties, resetprofile2_out_props)
        self.assertEqual(resetprofile2.manager, cpc1.reset_activation_profiles)

        # Because we know that the image and load profile managers are of the
        # same class, we don't need to test them.

    def test_profiles_remove(self):
        """Test remove() of FakedActivationProfileManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        resetprofiles = cpc1.reset_activation_profiles.list()
        resetprofile1 = resetprofiles[0]
        self.assertEqual(len(resetprofiles), 1)

        # the function to be tested:
        cpc1.reset_activation_profiles.remove(resetprofile1.oid)

        resetprofiles = cpc1.reset_activation_profiles.list()
        self.assertEqual(len(resetprofiles), 0)

        # Because we know that the image and load profile managers are of the
        # same class, we don't need to test them.


class FakedAdapterTests(unittest.TestCase):
    """All tests for the FakedAdapterManager and FakedAdapter classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.adapter1_in_props = {'name': 'adapter1', 'type': 'roce'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'adapters': [
                        {'properties': self.adapter1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedAdapterManager:
        self.hmc.add_resources(rd)

    def test_adapters_attr(self):
        """Test CPC 'adapters' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        self.assertIsInstance(cpc1.adapters, FakedAdapterManager)
        self.assertRegexpMatches(cpc1.adapters.base_uri, r'/api/adapters')

    def test_adapters_list(self):
        """Test list() of FakedAdapterManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        adapters = cpc1.adapters.list()

        self.assertEqual(len(adapters), 1)
        adapter1 = adapters[0]
        adapter1_out_props = self.adapter1_in_props.copy()
        adapter1_out_props.update({
            'object-id': adapter1.oid,
            'object-uri': adapter1.uri,
            'status': 'active',
            'adapter-family': 'roce',
            'network-port-uris': [],
        })
        self.assertIsInstance(adapter1, FakedAdapter)
        self.assertEqual(adapter1.properties, adapter1_out_props)
        self.assertEqual(adapter1.manager, cpc1.adapters)

        # Quick check of child resources:
        self.assertIsInstance(adapter1.ports, FakedPortManager)

    def test_adapters_add(self):
        """Test add() of FakedAdapterManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        self.assertEqual(len(adapters), 1)

        adapter2_in_props = {'name': 'adapter2', 'adapter-family': 'ficon'}

        # the function to be tested:
        new_adapter = cpc1.adapters.add(
            adapter2_in_props)

        adapters = cpc1.adapters.list()
        self.assertEqual(len(adapters), 2)

        adapter2 = [a for a in adapters
                    if a.properties['name'] == adapter2_in_props['name']][0]

        self.assertEqual(new_adapter.properties, adapter2.properties)
        self.assertEqual(new_adapter.manager, adapter2.manager)

        adapter2_out_props = adapter2_in_props.copy()
        adapter2_out_props.update({
            'object-id': adapter2.oid,
            'object-uri': adapter2.uri,
            'status': 'active',
            'storage-port-uris': [],
        })
        self.assertIsInstance(adapter2, FakedAdapter)
        self.assertEqual(adapter2.properties, adapter2_out_props)
        self.assertEqual(adapter2.manager, cpc1.adapters)

    def test_adapters_remove(self):
        """Test remove() of FakedAdapterManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]
        self.assertEqual(len(adapters), 1)

        # the function to be tested:
        cpc1.adapters.remove(adapter1.oid)

        adapters = cpc1.adapters.list()
        self.assertEqual(len(adapters), 0)


class FakedCpcTests(unittest.TestCase):
    """All tests for the FakedCpcManager and FakedCpc classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                },
            ]
        }
        # This already uses add() of FakedCpcManager:
        self.hmc.add_resources(rd)

    def test_cpcs_attr(self):
        """Test HMC 'cpcs' attribute."""
        self.assertIsInstance(self.hmc.cpcs, FakedCpcManager)
        self.assertRegexpMatches(self.hmc.cpcs.base_uri, r'/api/cpcs')

    def test_cpcs_list(self):
        """Test list() of FakedCpcManager."""

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        cpc1_out_props = self.cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })
        self.assertIsInstance(cpc1, FakedCpc)
        self.assertEqual(cpc1.properties, cpc1_out_props)
        self.assertEqual(cpc1.manager, self.hmc.cpcs)

        # Quick check of child resources:
        self.assertIsInstance(cpc1.lpars, FakedLparManager)
        self.assertIsInstance(cpc1.partitions, FakedPartitionManager)
        self.assertIsInstance(cpc1.adapters, FakedAdapterManager)
        self.assertIsInstance(cpc1.virtual_switches, FakedVirtualSwitchManager)
        self.assertIsInstance(cpc1.reset_activation_profiles,
                              FakedActivationProfileManager)
        self.assertIsInstance(cpc1.image_activation_profiles,
                              FakedActivationProfileManager)
        self.assertIsInstance(cpc1.load_activation_profiles,
                              FakedActivationProfileManager)

    def test_cpcs_add(self):
        """Test add() of FakedCpcManager."""
        cpcs = self.hmc.cpcs.list()
        self.assertEqual(len(cpcs), 1)

        cpc2_in_props = {'name': 'cpc2'}

        # the function to be tested:
        new_cpc = self.hmc.cpcs.add(cpc2_in_props)

        cpcs = self.hmc.cpcs.list()
        self.assertEqual(len(cpcs), 2)

        cpc2 = [cpc for cpc in cpcs
                if cpc.properties['name'] == cpc2_in_props['name']][0]

        self.assertEqual(new_cpc.properties, cpc2.properties)
        self.assertEqual(new_cpc.manager, cpc2.manager)

        cpc2_out_props = cpc2_in_props.copy()
        cpc2_out_props.update({
            'object-id': cpc2.oid,
            'object-uri': cpc2.uri,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })
        self.assertIsInstance(cpc2, FakedCpc)
        self.assertEqual(cpc2.properties, cpc2_out_props)
        self.assertEqual(cpc2.manager, self.hmc.cpcs)

    def test_cpcs_remove(self):
        """Test remove() of FakedCpcManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        self.assertEqual(len(cpcs), 1)

        # the function to be tested:
        self.hmc.cpcs.remove(cpc1.oid)

        cpcs = self.hmc.cpcs.list()
        self.assertEqual(len(cpcs), 0)


class FakedHbaTests(unittest.TestCase):
    """All tests for the FakedHbaManager and FakedHba classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')

        self.adapter1_oid = '747-abc-12345'
        self.adapter1_uri = '/api/adapters/%s' % self.adapter1_oid
        self.port1_oid = '23'
        self.port1_uri = '/api/adapters/%s/storage-ports/%s' % \
            (self.adapter1_oid, self.port1_oid)
        self.hba1_oid = '999-123-xyz'

        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        self.adapter1_in_props = {
            'object-id': self.adapter1_oid,
            'name': 'fcp1',
            'type': 'fcp',
        }
        self.port1_in_props = {
            'element-id': self.port1_oid,
            'name': 'port1',
        }
        self.hba1_in_props = {
            'element-id': self.hba1_oid,
            'name': 'hba1',
            'adapter-port-uri': self.port1_uri,
        }
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                            'hbas': [
                                {'properties': self.hba1_in_props},
                            ],
                        },
                    ],
                    'adapters': [
                        {
                            'properties': self.adapter1_in_props,
                            'ports': [
                                {'properties': self.port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedHbaManager:
        self.hmc.add_resources(rd)

    def test_hbas_attr(self):
        """Test Partition 'hbas' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        self.assertIsInstance(partition1.hbas, FakedHbaManager)
        self.assertRegexpMatches(partition1.hbas.base_uri,
                                 r'/api/partitions/[^/]+/hbas')

    def test_hbas_list(self):
        """Test list() of FakedHbaManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        # the function to be tested:
        hbas = partition1.hbas.list()

        self.assertEqual(len(hbas), 1)
        hba1 = hbas[0]
        hba1_out_props = self.hba1_in_props.copy()
        hba1_out_props.update({
            'element-id': self.hba1_oid,
            'element-uri': hba1.uri,
            'device-number': hba1.properties['device-number'],
            'wwpn': hba1.properties['wwpn'],
        })
        self.assertIsInstance(hba1, FakedHba)
        self.assertEqual(hba1.properties, hba1_out_props)
        self.assertEqual(hba1.manager, partition1.hbas)

    def test_hbas_add(self):
        """Test add() of FakedHbaManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        hbas = partition1.hbas.list()
        self.assertEqual(len(hbas), 1)

        hba2_oid = '22-55-xy'
        port_uri = '/api/adapters/abc-123/storage-ports/42'

        hba2_in_props = {
            'element-id': hba2_oid,
            'name': 'hba2',
            'adapter-port-uri': port_uri,
            'device-number': '8001',
            'wwpn': 'AFFEAFFE00008001',
        }

        # the function to be tested:
        new_hba = partition1.hbas.add(
            hba2_in_props)

        hbas = partition1.hbas.list()
        self.assertEqual(len(hbas), 2)

        hba2 = [hba for hba in hbas
                if hba.properties['name'] == hba2_in_props['name']][0]

        self.assertEqual(new_hba.properties, hba2.properties)
        self.assertEqual(new_hba.manager, hba2.manager)

        hba2_out_props = hba2_in_props.copy()
        hba2_out_props.update({
            'element-id': hba2_oid,
            'element-uri': hba2.uri,
        })
        self.assertIsInstance(hba2, FakedHba)
        self.assertEqual(hba2.properties, hba2_out_props)
        self.assertEqual(hba2.manager, partition1.hbas)

    def test_hbas_remove(self):
        """Test remove() of FakedHbaManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        hbas = partition1.hbas.list()
        hba1 = hbas[0]
        self.assertEqual(len(hbas), 1)

        # the function to be tested:
        partition1.hbas.remove(hba1.oid)

        hbas = partition1.hbas.list()
        self.assertEqual(len(hbas), 0)

    # TODO: Add testcases for updating 'hba-uris' parent property


class FakedLparTests(unittest.TestCase):
    """All tests for the FakedLparManager and FakedLpar classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.lpar1_in_props = {'name': 'lpar1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'lpars': [
                        {'properties': self.lpar1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedLparManager:
        self.hmc.add_resources(rd)

    def test_lpars_attr(self):
        """Test CPC 'lpars' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        self.assertIsInstance(cpc1.lpars, FakedLparManager)
        self.assertRegexpMatches(cpc1.lpars.base_uri,
                                 r'/api/logical-partitions')

    def test_lpars_list(self):
        """Test list() of FakedLparManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        lpars = cpc1.lpars.list()

        self.assertEqual(len(lpars), 1)
        lpar1 = lpars[0]
        lpar1_out_props = self.lpar1_in_props.copy()
        lpar1_out_props.update({
            'object-id': lpar1.oid,
            'object-uri': lpar1.uri,
            'status': 'not-activated',
        })
        self.assertIsInstance(lpar1, FakedLpar)
        self.assertEqual(lpar1.properties, lpar1_out_props)
        self.assertEqual(lpar1.manager, cpc1.lpars)

    def test_lpars_add(self):
        """Test add() of FakedLparManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        lpars = cpc1.lpars.list()
        self.assertEqual(len(lpars), 1)

        lpar2_in_props = {'name': 'lpar2'}

        # the function to be tested:
        new_lpar = cpc1.lpars.add(
            lpar2_in_props)

        lpars = cpc1.lpars.list()
        self.assertEqual(len(lpars), 2)

        lpar2 = [p for p in lpars
                 if p.properties['name'] == lpar2_in_props['name']][0]

        self.assertEqual(new_lpar.properties, lpar2.properties)
        self.assertEqual(new_lpar.manager, lpar2.manager)

        lpar2_out_props = lpar2_in_props.copy()
        lpar2_out_props.update({
            'object-id': lpar2.oid,
            'object-uri': lpar2.uri,
            'status': 'not-activated',
        })
        self.assertIsInstance(lpar2, FakedLpar)
        self.assertEqual(lpar2.properties, lpar2_out_props)
        self.assertEqual(lpar2.manager, cpc1.lpars)

    def test_lpars_remove(self):
        """Test remove() of FakedLparManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        lpars = cpc1.lpars.list()
        lpar1 = lpars[0]
        self.assertEqual(len(lpars), 1)

        # the function to be tested:
        cpc1.lpars.remove(lpar1.oid)

        lpars = cpc1.lpars.list()
        self.assertEqual(len(lpars), 0)


class FakedNicTests(unittest.TestCase):
    """All tests for the FakedNicManager and FakedNic classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')

        self.adapter1_oid = '380-xyz-12345'
        self.adapter1_uri = '/api/adapters/%s' % self.adapter1_oid
        self.port1_oid = '32'
        self.port1_uri = '/api/adapters/%s/network-ports/%s' % \
            (self.adapter1_oid, self.port1_oid)
        self.nic1_oid = 'ddd-999-123'

        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        self.nic1_in_props = {
            'element-id': self.nic1_oid,
            'name': 'nic1',
            'network-adapter-port-uri': self.port1_uri,
        }
        self.adapter1_in_props = {
            'object-id': self.adapter1_oid,
            'name': 'roce1',
            'type': 'roce',
        }
        self.port1_in_props = {
            'element-id': self.port1_oid,
            'name': 'port1',
        }
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                            'nics': [
                                {'properties': self.nic1_in_props},
                            ],
                        },
                    ],
                    'adapters': [
                        {
                            'properties': self.adapter1_in_props,
                            'ports': [
                                {'properties': self.port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedNicManager:
        self.hmc.add_resources(rd)

    def test_nics_attr(self):
        """Test Partition 'nics' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        self.assertIsInstance(partition1.nics, FakedNicManager)
        self.assertRegexpMatches(partition1.nics.base_uri,
                                 r'/api/partitions/[^/]+/nics')

    def test_nics_list(self):
        """Test list() of FakedNicManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        # the function to be tested:
        nics = partition1.nics.list()

        self.assertEqual(len(nics), 1)
        nic1 = nics[0]
        nic1_out_props = self.nic1_in_props.copy()
        nic1_out_props.update({
            'element-id': self.nic1_oid,
            'element-uri': nic1.uri,
            'device-number': nic1.properties['device-number'],
        })
        self.assertIsInstance(nic1, FakedNic)
        self.assertEqual(nic1.properties, nic1_out_props)
        self.assertEqual(nic1.manager, partition1.nics)

    def test_nics_add(self):
        """Test add() of FakedNicManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        nics = partition1.nics.list()
        self.assertEqual(len(nics), 1)

        nic2_oid = '77-55-ab'
        port_uri = '/api/adapters/abc-123/network-ports/42'

        nic2_in_props = {
            'element-id': nic2_oid,
            'name': 'nic2',
            'network-adapter-port-uri': port_uri,
        }

        # the function to be tested:
        new_nic = partition1.nics.add(
            nic2_in_props)

        nics = partition1.nics.list()
        self.assertEqual(len(nics), 2)

        nic2 = [nic for nic in nics
                if nic.properties['name'] == nic2_in_props['name']][0]

        self.assertEqual(new_nic.properties, nic2.properties)
        self.assertEqual(new_nic.manager, nic2.manager)

        nic2_out_props = nic2_in_props.copy()
        nic2_out_props.update({
            'element-id': nic2_oid,
            'element-uri': nic2.uri,
            'device-number': nic2.properties['device-number'],
        })
        self.assertIsInstance(nic2, FakedNic)
        self.assertEqual(nic2.properties, nic2_out_props)
        self.assertEqual(nic2.manager, partition1.nics)

    def test_nics_remove(self):
        """Test remove() of FakedNicManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        nics = partition1.nics.list()
        nic1 = nics[0]
        self.assertEqual(len(nics), 1)

        # the function to be tested:
        partition1.nics.remove(nic1.oid)

        nics = partition1.nics.list()
        self.assertEqual(len(nics), 0)

    # TODO: Add testcases for updating 'nic-uris' parent property


class FakedPartitionTests(unittest.TestCase):
    """All tests for the FakedPartitionManager and FakedPartition classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {'properties': self.partition1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedPartitionManager:
        self.hmc.add_resources(rd)

    def test_partitions_attr(self):
        """Test CPC 'partitions' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        self.assertIsInstance(cpc1.partitions, FakedPartitionManager)
        self.assertRegexpMatches(cpc1.partitions.base_uri, r'/api/partitions')

    def test_partitions_list(self):
        """Test list() of FakedPartitionManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        partitions = cpc1.partitions.list()

        self.assertEqual(len(partitions), 1)
        partition1 = partitions[0]
        partition1_out_props = self.partition1_in_props.copy()
        partition1_out_props.update({
            'object-id': partition1.oid,
            'object-uri': partition1.uri,
            'status': 'stopped',
            'hba-uris': [],
            'nic-uris': [],
            'virtual-function-uris': [],
        })
        self.assertIsInstance(partition1, FakedPartition)
        self.assertEqual(partition1.properties, partition1_out_props)
        self.assertEqual(partition1.manager, cpc1.partitions)

    def test_partitions_add(self):
        """Test add() of FakedPartitionManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        self.assertEqual(len(partitions), 1)

        partition2_in_props = {'name': 'partition2'}

        # the function to be tested:
        new_partition = cpc1.partitions.add(
            partition2_in_props)

        partitions = cpc1.partitions.list()
        self.assertEqual(len(partitions), 2)

        partition2 = [p for p in partitions
                      if p.properties['name'] ==
                      partition2_in_props['name']][0]

        self.assertEqual(new_partition.properties, partition2.properties)
        self.assertEqual(new_partition.manager, partition2.manager)

        partition2_out_props = partition2_in_props.copy()
        partition2_out_props.update({
            'object-id': partition2.oid,
            'object-uri': partition2.uri,
            'status': 'stopped',
            'hba-uris': [],
            'nic-uris': [],
            'virtual-function-uris': [],
        })
        self.assertIsInstance(partition2, FakedPartition)
        self.assertEqual(partition2.properties, partition2_out_props)
        self.assertEqual(partition2.manager, cpc1.partitions)

    def test_partitions_remove(self):
        """Test remove() of FakedPartitionManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        self.assertEqual(len(partitions), 1)

        # the function to be tested:
        cpc1.partitions.remove(partition1.oid)

        partitions = cpc1.partitions.list()
        self.assertEqual(len(partitions), 0)


class FakedPortTests(unittest.TestCase):
    """All tests for the FakedPortManager and FakedPort classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.adapter1_in_props = {'name': 'adapter1', 'adapter-family': 'osa'}
        self.port1_in_props = {'name': 'port1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'adapters': [
                        {
                            'properties': self.adapter1_in_props,
                            'ports': [
                                {'properties': self.port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedPortManager:
        self.hmc.add_resources(rd)

    def test_ports_attr(self):
        """Test Adapter 'ports' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]

        self.assertIsInstance(adapter1.ports, FakedPortManager)
        self.assertRegexpMatches(adapter1.ports.base_uri,
                                 r'/api/adapters/[^/]+/network-ports')

    def test_ports_list(self):
        """Test list() of FakedPortManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]

        # the function to be tested:
        ports = adapter1.ports.list()

        self.assertEqual(len(ports), 1)
        port1 = ports[0]
        port1_out_props = self.port1_in_props.copy()
        port1_out_props.update({
            'element-id': port1.oid,
            'element-uri': port1.uri,
        })
        self.assertIsInstance(port1, FakedPort)
        self.assertEqual(port1.properties, port1_out_props)
        self.assertEqual(port1.manager, adapter1.ports)

    def test_ports_add(self):
        """Test add() of FakedPortManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]
        ports = adapter1.ports.list()
        self.assertEqual(len(ports), 1)

        port2_in_props = {'name': 'port2'}

        # the function to be tested:
        new_port = adapter1.ports.add(
            port2_in_props)

        ports = adapter1.ports.list()
        self.assertEqual(len(ports), 2)

        port2 = [p for p in ports
                 if p.properties['name'] == port2_in_props['name']][0]

        self.assertEqual(new_port.properties, port2.properties)
        self.assertEqual(new_port.manager, port2.manager)

        port2_out_props = port2_in_props.copy()
        port2_out_props.update({
            'element-id': port2.oid,
            'element-uri': port2.uri,
        })
        self.assertIsInstance(port2, FakedPort)
        self.assertEqual(port2.properties, port2_out_props)
        self.assertEqual(port2.manager, adapter1.ports)

    def test_ports_remove(self):
        """Test remove() of FakedPortManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]
        ports = adapter1.ports.list()
        port1 = ports[0]
        self.assertEqual(len(ports), 1)

        # the function to be tested:
        adapter1.ports.remove(port1.oid)

        ports = adapter1.ports.list()
        self.assertEqual(len(ports), 0)

    # TODO: Add testcases for updating 'network-port-uris' and
    #       'storage-port-uris' parent properties


class FakedVirtualFunctionTests(unittest.TestCase):
    """All tests for the FakedVirtualFunctionManager and FakedVirtualFunction
    classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        self.virtual_function1_in_props = {'name': 'virtual_function1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                            'virtual_functions': [
                                {'properties':
                                 self.virtual_function1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedVirtualFunctionManager:
        self.hmc.add_resources(rd)

    def test_virtual_functions_attr(self):
        """Test CPC 'virtual_functions' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        self.assertIsInstance(partition1.virtual_functions,
                              FakedVirtualFunctionManager)
        self.assertRegexpMatches(partition1.virtual_functions.base_uri,
                                 r'/api/partitions/[^/]+/virtual-functions')

    def test_virtual_functions_list(self):
        """Test list() of FakedVirtualFunctionManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        # the function to be tested:
        virtual_functions = partition1.virtual_functions.list()

        self.assertEqual(len(virtual_functions), 1)
        virtual_function1 = virtual_functions[0]
        virtual_function1_out_props = self.virtual_function1_in_props.copy()
        virtual_function1_out_props.update({
            'element-id': virtual_function1.oid,
            'element-uri': virtual_function1.uri,
            'device-number': virtual_function1.properties['device-number'],
        })
        self.assertIsInstance(virtual_function1, FakedVirtualFunction)
        self.assertEqual(virtual_function1.properties,
                         virtual_function1_out_props)
        self.assertEqual(virtual_function1.manager,
                         partition1.virtual_functions)

    def test_virtual_functions_add(self):
        """Test add() of FakedVirtualFunctionManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        virtual_functions = partition1.virtual_functions.list()
        self.assertEqual(len(virtual_functions), 1)

        virtual_function2_in_props = {'name': 'virtual_function2'}

        # the function to be tested:
        new_virtual_function = partition1.virtual_functions.add(
            virtual_function2_in_props)

        virtual_functions = partition1.virtual_functions.list()
        self.assertEqual(len(virtual_functions), 2)

        virtual_function2 = [vf for vf in virtual_functions
                             if vf.properties['name'] ==
                             virtual_function2_in_props['name']][0]

        self.assertEqual(new_virtual_function.properties,
                         virtual_function2.properties)
        self.assertEqual(new_virtual_function.manager,
                         virtual_function2.manager)

        virtual_function2_out_props = virtual_function2_in_props.copy()
        virtual_function2_out_props.update({
            'element-id': virtual_function2.oid,
            'element-uri': virtual_function2.uri,
            'device-number': virtual_function2.properties['device-number'],
        })
        self.assertIsInstance(virtual_function2, FakedVirtualFunction)
        self.assertEqual(virtual_function2.properties,
                         virtual_function2_out_props)
        self.assertEqual(virtual_function2.manager,
                         partition1.virtual_functions)

    def test_virtual_functions_remove(self):
        """Test remove() of FakedVirtualFunctionManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        virtual_functions = partition1.virtual_functions.list()
        virtual_function1 = virtual_functions[0]
        self.assertEqual(len(virtual_functions), 1)

        # the function to be tested:
        partition1.virtual_functions.remove(virtual_function1.oid)

        virtual_functions = partition1.virtual_functions.list()
        self.assertEqual(len(virtual_functions), 0)

    # TODO: Add testcases for updating 'virtual-function-uris' parent property


class FakedVirtualSwitchTests(unittest.TestCase):
    """All tests for the FakedVirtualSwitchManager and FakedVirtualSwitch
    classes."""

    def setUp(self):
        self.hmc = FakedHmc('fake-hmc', '2.13.1', '1.8')
        self.cpc1_in_props = {'name': 'cpc1'}
        self.virtual_switch1_in_props = {'name': 'virtual_switch1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'virtual_switches': [
                        {'properties': self.virtual_switch1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedVirtualSwitchManager:
        self.hmc.add_resources(rd)

    def test_virtual_switches_attr(self):
        """Test CPC 'virtual_switches' attribute."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        self.assertIsInstance(cpc1.virtual_switches, FakedVirtualSwitchManager)
        self.assertRegexpMatches(cpc1.virtual_switches.base_uri,
                                 r'/api/virtual-switches')

    def test_virtual_switches_list(self):
        """Test list() of FakedVirtualSwitchManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        virtual_switches = cpc1.virtual_switches.list()

        self.assertEqual(len(virtual_switches), 1)
        virtual_switch1 = virtual_switches[0]
        virtual_switch1_out_props = self.virtual_switch1_in_props.copy()
        virtual_switch1_out_props.update({
            'object-id': virtual_switch1.oid,
            'object-uri': virtual_switch1.uri,
        })
        self.assertIsInstance(virtual_switch1, FakedVirtualSwitch)
        self.assertEqual(virtual_switch1.properties, virtual_switch1_out_props)
        self.assertEqual(virtual_switch1.manager, cpc1.virtual_switches)

    def test_virtual_switches_add(self):
        """Test add() of FakedVirtualSwitchManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        virtual_switches = cpc1.virtual_switches.list()
        self.assertEqual(len(virtual_switches), 1)

        virtual_switch2_in_props = {'name': 'virtual_switch2'}

        # the function to be tested:
        new_virtual_switch = cpc1.virtual_switches.add(
            virtual_switch2_in_props)

        virtual_switches = cpc1.virtual_switches.list()
        self.assertEqual(len(virtual_switches), 2)

        virtual_switch2 = [p for p in virtual_switches
                           if p.properties['name'] ==
                           virtual_switch2_in_props['name']][0]

        self.assertEqual(new_virtual_switch.properties,
                         virtual_switch2.properties)
        self.assertEqual(new_virtual_switch.manager,
                         virtual_switch2.manager)

        virtual_switch2_out_props = virtual_switch2_in_props.copy()
        virtual_switch2_out_props.update({
            'object-id': virtual_switch2.oid,
            'object-uri': virtual_switch2.uri,
        })
        self.assertIsInstance(virtual_switch2, FakedVirtualSwitch)
        self.assertEqual(virtual_switch2.properties, virtual_switch2_out_props)
        self.assertEqual(virtual_switch2.manager, cpc1.virtual_switches)

    def test_virtual_switches_remove(self):
        """Test remove() of FakedVirtualSwitchManager."""
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        virtual_switches = cpc1.virtual_switches.list()
        virtual_switch1 = virtual_switches[0]
        self.assertEqual(len(virtual_switches), 1)

        # the function to be tested:
        cpc1.virtual_switches.remove(virtual_switch1.oid)

        virtual_switches = cpc1.virtual_switches.list()
        self.assertEqual(len(virtual_switches), 0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example unit test for a user of the zhmcclient package.
"""

from __future__ import absolute_import, print_function

import requests.packages.urllib3
import unittest
import zhmcclient
import zhmcclient_mock


class MyTests(unittest.TestCase):

    @staticmethod
    def create_session_1():
        """
        Demonstrate how to populate a faked session with resources defined
        in a resource dictionary.
        """
        session = zhmcclient_mock.FakedSession('fake-host', 'fake-hmc',
                                               '2.13.1', '1.8')
        session.hmc.add_resources({
            'cpcs': [
                {
                    'properties': {
                        # object-id is auto-generated
                        # object-uri is auto-generated
                        'name': 'cpc_1',
                        'dpm-enabled': False,
                        'description': 'CPC #1',
                    },
                    'lpars': [
                        {
                            'properties': {
                                # object-id is auto-generated
                                # object-uri is auto-generated
                                'name': 'lpar_1',
                                'description': 'LPAR #1 in CPC #1',
                            },
                        },
                    ],
                },
                {
                    'properties': {
                        # object-id is auto-generated
                        # object-uri is auto-generated
                        'name': 'cpc_2',
                        'dpm-enabled': True,
                        'description': 'CPC #2',
                    },
                    'partitions': [
                        {
                            'properties': {
                                # object-id is auto-generated
                                # object-uri is auto-generated
                                'name': 'partition_1',
                                'description': 'Partition #1 in CPC #2',
                            },
                        },
                    ],
                    'adapters': [
                        {
                            'properties': {
                                # object-id is auto-generated
                                # object-uri is auto-generated
                                'name': 'osa_1',
                                'description': 'OSA #1 in CPC #2',
                                'type': 'osd',
                            },
                            'ports': [
                                {
                                    'properties': {
                                        # element-id is auto-generated
                                        # element-uri is auto-generated
                                        'name': 'osa_1_port_1',
                                        'description': 'Port #1 of OSA #1',
                                    },
                                },
                            ],
                        },
                    ],
                },
            ],
        })
        return session

    @staticmethod
    def create_session_2():
        """
        Demonstrate how to populate a faked session with resources one by one.
        """
        session = zhmcclient_mock.FakedSession('fake-host', 'fake-hmc',
                                               '2.13.1', '1.8')
        cpc1 = session.hmc.cpcs.add({
            # object-id is auto-generated
            # object-uri is auto-generated
            'name': 'cpc_1',
            'dpm-enabled': False,
            'description': 'CPC #1',
        })
        cpc1.lpars.add({
            # object-id is auto-generated
            # object-uri is auto-generated
            'name': 'lpar_1',
            'description': 'LPAR #1 in CPC #1',
        })
        cpc2 = session.hmc.cpcs.add({
            # object-id is auto-generated
            # object-uri is auto-generated
            'name': 'cpc_2',
            'dpm-enabled': True,
            'description': 'CPC #2',
        })
        cpc2.partitions.add({
            # object-id is auto-generated
            # object-uri is auto-generated
            'name': 'partition_1',
            'description': 'Partition #1 in CPC #2',
        })
        adapter1 = cpc2.adapters.add({
            # object-id is auto-generated
            # object-uri is auto-generated
            'name': 'osa_1',
            'description': 'OSA #1 in CPC #2',
            'type': 'osd',
        })
        adapter1.ports.add({
            # element-id is auto-generated
            # element-uri is auto-generated
            'name': 'osa_1_port_1',
            'description': 'Port #1 of OSA #1',
        })
        return session

    def check(self):
        """
        Check the faked session and its faked HMC.
        """

        self.assertEqual(self.session.host, 'fake-host')

        self.assertEqual(self.client.version_info(), (1, 8))

        cpcs = self.client.cpcs.list()
        self.assertEqual(len(cpcs), 2)

        cpc1 = cpcs[0]  # a CPC in classic mode
        self.assertEqual(cpc1.get_property('name'), 'cpc_1')
        self.assertFalse(cpc1.dpm_enabled)

        lpars = cpc1.lpars.list()
        self.assertEqual(len(lpars), 1)
        lpar1 = lpars[0]
        self.assertEqual(lpar1.get_property('name'), 'lpar_1')

        cpc2 = cpcs[1]  # a CPC in DPM mode
        self.assertEqual(cpc2.get_property('name'), 'cpc_2')
        self.assertTrue(cpc2.dpm_enabled)

        partitions = cpc2.partitions.list()
        self.assertEqual(len(partitions), 1)
        partition1 = partitions[0]
        self.assertEqual(partition1.get_property('name'), 'partition_1')

        adapters = cpc2.adapters.list()
        self.assertEqual(len(adapters), 1)
        adapter1 = adapters[0]
        self.assertEqual(adapter1.get_property('name'), 'osa_1')

        ports = adapter1.ports.list()
        self.assertEqual(len(ports), 1)
        port1 = ports[0]
        self.assertEqual(port1.get_property('name'), 'osa_1_port_1')

    def test_session_1(self):
        self.session = self.create_session_1()
        self.client = zhmcclient.Client(self.session)
        self.check()

    def test_session_2(self):
        self.session = self.create_session_2()
        self.client = zhmcclient.Client(self.session)
        self.check()


if __name__ == '__main__':
    requests.packages.urllib3.disable_warnings()
    unittest.main()



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR
from ._cmd_cpc import find_cpc


def find_lpar(client, cpc_name, lpar_name):
    """
    Find an LPAR by name and return its resource object.
    """
    cpc = find_cpc(client, cpc_name)
    # The CPC must not be in DPM mode. We don't check that because it would
    # cause a GET to the CPC resource that we otherwise don't need.
    try:
        lpar = cpc.lpars.find(name=lpar_name)
    except zhmcclient.NotFound:
        raise click.ClickException("Could not find LPAR %s in CPC %s." %
                                   (lpar_name, cpc_name))
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))
    return lpar


@cli.group('lpar', options_metavar=COMMAND_OPTIONS_METAVAR)
def lpar_group():
    """
    Command group for managing LPARs.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@lpar_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--type', is_flag=True, required=False,
              help='Show additional properties indicating the LPAR and OS '
              'type.')
@click.option('--uri', is_flag=True, required=False,
              help='Show additional properties for the resource URI.')
@click.pass_obj
def lpar_list(cmd_ctx, cpc, **options):
    """
    List the LPARs in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_lpar_list(cmd_ctx, cpc, options))


@lpar_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('LPAR', type=str, metavar='LPAR')
@click.pass_obj
def lpar_show(cmd_ctx, cpc, lpar):
    """
    Show details of an LPAR in a CPC.

    \b
    Limitations:
      * In table format, the following properties are not shown:
        - program-status-word-information

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_lpar_show(cmd_ctx, cpc, lpar))


@lpar_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('LPAR', type=str, metavar='LPAR')
@click.option('--acceptable-status', type=str, required=False,
              help='The new set of acceptable operational status values.')
# TODO: Support multiple values for acceptable-status
@click.option('--next-activation-profile', type=str, required=False,
              help='The name of the new next image or load activation '
              'profile.')
# TODO: Add support for updating processor capping/sharing/weight related props
@click.option('--zaware-host-name', type=str, required=False,
              help='The new hostname for IBM zAware. '
              '(only for LPARs in zaware activation mode).')
@click.option('--zaware-master-userid', type=str, required=False,
              help='The new master userid for IBM zAware. '
              '(only for LPARs in zaware activation mode).')
@click.option('--zaware-master-password', type=str, required=False,
              help='The new master password for IBM zAware. '
              '(only for LPARs in zaware activation mode).')
# TODO: Change zAware master password option to ask for password
# TODO: Add support for updating zAware network-related properties
@click.option('--ssc-host-name', type=str, required=False,
              help='The new hostname for the SSC appliance. '
              '(only for LPARs in ssc activation mode).')
@click.option('--ssc-master-userid', type=str, required=False,
              help='The new master userid for the SSC appliance. '
              '(only for LPARs in ssc activation mode).')
@click.option('--ssc-master-password', type=str, required=False,
              help='The new master password for the SSC appliance. '
              '(only for LPARs in ssc activation mode).')
# TODO: Change SSC master password option to ask for password
# TODO: Add support for updating SSC network-related properties
@click.pass_obj
def lpar_update(cmd_ctx, cpc, lpar, **options):
    """
    Update the properties of an LPAR.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.

    \b
    Limitations:
      * The --acceptable-status option does not support multiple values.
      * The processor capping/sharing/weight related properties cannot be
        updated.
      * The network-related properties for zaware and ssc cannot beupdated.
      * The --zaware-master-password and --ssc-master-password options do not
        ask for the password.
    """
    cmd_ctx.execute_cmd(lambda: cmd_lpar_update(cmd_ctx, cpc, lpar, options))


@lpar_group.command('activate', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('LPAR', type=str, metavar='LPAR')
@click.pass_obj
def lpar_activate(cmd_ctx, cpc, lpar):
    """
    Activate an LPAR.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_lpar_activate(cmd_ctx, cpc, lpar))


@lpar_group.command('deactivate', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('LPAR', type=str, metavar='LPAR')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deactivation of the LPAR.',
              prompt='Are you sure you want to deactivate the LPAR ?')
@click.pass_obj
def lpar_deactivate(cmd_ctx, cpc, lpar):
    """
    Deactivate an LPAR.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_lpar_deactivate(cmd_ctx, cpc, lpar))


@lpar_group.command('load', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('LPAR', type=str, metavar='LPAR')
@click.argument('LOAD-ADDRESS', type=str, metavar='LOAD-ADDRESS')
@click.option('--load-parameter', type=str, required=False,
              help='Provides additional control over the outcome of a '
              'Load operation.')
@click.pass_obj
def lpar_load(cmd_ctx, cpc, lpar, load_address, **options):
    """
    Load (Boot, IML) an LPAR.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_lpar_load(cmd_ctx, cpc, lpar,
                                              load_address, options))


def cmd_lpar_list(cmd_ctx, cpc_name, options):

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(client, cpc_name)

    try:
        lpars = cpc.lpars.list()
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))

    show_list = [
        'name',
        'status',
    ]
    if options['type']:
        show_list.extend([
            'activation-mode',
            'os-type',
            'workload-manager-enabled',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    cmd_ctx.spinner.stop()
    print_resources(lpars, cmd_ctx.output_format, show_list)


def cmd_lpar_show(cmd_ctx, cpc_name, lpar_name):

    client = zhmcclient.Client(cmd_ctx.session)
    lpar = find_lpar(client, cpc_name, lpar_name)

    try:
        lpar.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))

    skip_list = (
        'program-status-word-information',
    )

    cmd_ctx.spinner.stop()
    print_properties(lpar.properties, cmd_ctx.output_format, skip_list)


def cmd_lpar_update(cmd_ctx, cpc_name, lpar_name, options):

    client = zhmcclient.Client(cmd_ctx.session)
    lpar = find_lpar(client, cpc_name, lpar_name)

    name_map = {
        'next-activation-profile': 'next-activation-profile-name',
    }
    options = original_options(options)
    properties = options_to_properties(options, name_map)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating LPAR %s." % lpar_name)
        return

    try:
        lpar.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))

    cmd_ctx.spinner.stop()
    # LPARs cannot be renamed.
    click.echo("LPAR %s has been updated." % lpar_name)


def cmd_lpar_activate(cmd_ctx, cpc_name, lpar_name):

    client = zhmcclient.Client(cmd_ctx.session)
    lpar = find_lpar(client, cpc_name, lpar_name)

    try:
        lpar.activate(wait_for_completion=True)
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))

    cmd_ctx.spinner.stop()
    click.echo('Activation of LPAR %s is complete.' % lpar_name)


def cmd_lpar_deactivate(cmd_ctx, cpc_name, lpar_name):

    client = zhmcclient.Client(cmd_ctx.session)
    lpar = find_lpar(client, cpc_name, lpar_name)

    try:
        lpar.deactivate(wait_for_completion=True)
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))

    cmd_ctx.spinner.stop()
    click.echo('Deactivation of LPAR %s is complete.' % lpar_name)


def cmd_lpar_load(cmd_ctx, cpc_name, lpar_name, load_address, options):

    client = zhmcclient.Client(cmd_ctx.session)
    lpar = find_lpar(client, cpc_name, lpar_name)

    options = original_options(options)

    try:
        lpar.load(load_address, wait_for_completion=True, **options)
    except zhmcclient.Error as exc:
        raise click.ClickException("%s: %s" % (exc.__class__.__name__, exc))

    cmd_ctx.spinner.stop()
    click.echo('Loading of LPAR %s is complete.' % lpar_name)



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The :class:`~zhmcclient.TimeStatsKeeper` class allows measuring the elapsed
time of accordingly instrumented code and keeps a statistics of these times.

The :class:`~zhmcclient.Session` class uses this class for keeping
statistics about the time to issue HTTP requests against the HMC API (see its
:attr:`~zhmcclient.Session.time_stats_keeper` property).

The :class:`~zhmcclient.TimeStats` class is a helper class that contains the
actual measurement data for all invocations of a particular HTTP request. Its
objects are under control of the :class:`~zhmcclient.TimeStatsKeeper` class.

Example:

::

    import zhmcclient

    session = zhmcclient.Session(hmc, userid, password)
    session.time_stats_keeper.enable()

    # Some operations that are being measured
    client = zhmcclient.Client(session)
    cpcs = client.cpcs.list()

    print(session.time_stats_keeper)
"""

from __future__ import absolute_import

import time
import copy

from ._logging import get_logger, logged_api_call

__all__ = ['TimeStatsKeeper', 'TimeStats']

LOG = get_logger(__name__)


class TimeStats(object):
    """
    Elapsed time statistics for all invocations of a particular named
    operation.

    All invocations of the operation will be accumulated into the statistics
    data kept by an object of this class.

    Objects of this class don't need to (and in fact, are not supposed to) be
    created by the user. Instead, the
    :meth:`zhmcclient.TimeStatsKeeper.get_stats` method should be used to
    create objects of this class.
    """

    def __init__(self, keeper, name):
        """
        Parameters:

          keeper (TimeStatsKeeper):
            The statistics keeper that holds this time statistics.

          name (string):
            Name of the operation.
        """
        self._keeper = keeper
        self._name = name
        self._count = 0
        self._sum = float(0)
        self._min = float('inf')
        self._max = float(0)
        self._begin_time = None

    @property
    def name(self):
        """
        :term:`string`: Name of the operation this time statistics has data
        for.

        This name is used by the :class:`~zhmcclient.TimeStatsKeeper` object
        holding this time statistics as a key.
        """
        return self._name

    @property
    def keeper(self):
        """
        :class:`~zhmcclient.TimeStatsKeeper`: The time statistics keeper
        holding this time statistics.
        """
        return self._keeper

    @property
    def count(self):
        """
        :term:`integer`: The number of invocations of the operation.
        """
        return self._count

    @property
    def avg_time(self):
        """
        float: The average elapsed time for invoking the operation, in seconds.
        """
        try:
            return self._sum / self._count
        except ZeroDivisionError:
            return 0

    @property
    def min_time(self):
        """
        float: The minimum elapsed time for invoking the operation, in seconds.
        """
        return self._min

    @property
    def max_time(self):
        """
        float: The maximum elapsed time for invoking the operation, in seconds.
        """
        return self._max

    @logged_api_call
    def reset(self):
        """
        Reset the time statistics data for the operation.
        """
        self._count = 0
        self._sum = float(0)
        self._min = float('inf')
        self._max = float(0)

    @logged_api_call
    def begin(self):
        """
        This method must be called before invoking the operation.
        Note that this method is not to be invoked by the user; it is invoked
        by the implementation of the :class:`~zhmcclient.Session` class.

        If the statistics keeper holding this time statistics is enabled, this
        method takes the current time, so that
        :meth:`~zhmcclient.TimeStats.end` can calculate the elapsed time
        between the two method calls.

        If the statistics keeper holding this time statistics is disabled,
        this method does nothing, in order to save resources.
        """
        if self.keeper.enabled:
            self._begin_time = time.time()

    @logged_api_call
    def end(self):
        """
        This method must be called after the operation returns.
        Note that this method is not to be invoked by the user; it is invoked
        by the implementation of the :class:`~zhmcclient.Session` class.

        If the statistics keeper holding this time statistics is enabled, this
        method takes the current time, calculates the duration of the operation
        since the last call to :meth:`~zhmcclient.TimeStats.begin`, and updates
        the time statistics to reflect the new operation.

        If the statistics keeper holding this time statistics is disabled,
        this method does nothing, in order to save resources.
        """
        if self.keeper.enabled:
            if self._begin_time is None:
                raise RuntimeError("end() called without preceding begin()")
            dt = time.time() - self._begin_time
            self._begin_time = None
            self._count += 1
            self._sum += dt
            if dt > self._max:
                self._max = dt
            if dt < self._min:
                self._min = dt

    def __str__(self):
        """
        Return a human readable string with the time statistics for this
        operation.

        Example result:

        ::

            TimeStats: count=1 avg=1.000s min=1.000s max=1.000s get /api/cpcs
        """
        return "TimeStats: count={:d} avg={:.3f}s min={:.3f}s "\
               "max={:.3f}s {}".format(
                   self.count, self.avg_time, self.min_time, self.max_time,
                   self.name)


class TimeStatsKeeper(object):
    """
    Statistics keeper for elapsed times.

    The statistics keeper can hold multiple time statistics (see
    :class:`~zhmcclient.TimeStats`), that are identified by a name.

    The statistics keeper can be in a state of enabled or disabled. If enabled,
    it accumulates the elapsed times between subsequent calls to the
    :meth:`~zhmcclient.TimeStats.begin` and :meth:`~zhmcclient.TimeStats.end`
    methods of class :class:`~zhmcclient.TimeStats`.
    If disabled, calls to these methods do not accumulate any time.

    Initially, the statistics keeper is disabled.
    """

    def __init__(self):
        self._enabled = False
        self._time_stats = {}  # TimeStats objects
        self._disabled_stats = TimeStats(self, "disabled")

    @property
    def enabled(self):
        """
        Indicates whether the statistics keeper is enabled.
        """
        return self._enabled

    @logged_api_call
    def enable(self):
        """
        Enable the statistics keeper.
        """
        self._enabled = True

    @logged_api_call
    def disable(self):
        """
        Disable the statistics keeper.
        """
        self._enabled = False

    @logged_api_call
    def get_stats(self, name):
        """
        Get the time statistics for a name.
        If a time statistics for that name does not exist yet, create one.

        Parameters:

          name (string):
            Name of the time statistics.

        Returns:

          TimeStats: The time statistics for the specified name. If the
          statistics keeper is disabled, a dummy time statistics object is
          returned, in order to save resources.
        """
        if not self.enabled:
            return self._disabled_stats
        if name not in self._time_stats:
            self._time_stats[name] = TimeStats(self, name)
        return self._time_stats[name]

    @logged_api_call
    def snapshot(self):
        """
        Return a snapshot of the time statistics of this keeper.

        The snapshot represents the statistics data at the time this method
        is called, and remains unchanged even if the statistics of this keeper
        continues to be updated.

        Returns:

          : A list of tuples (name, stats) with:

          - name (:term:`string`): Name of the operation
          - stats (:class:`~zhmcclient.TimeStats`): Time statistics for the
            operation
        """
        return copy.deepcopy(self._time_stats).items()

    def __str__(self):
        """
        Return a human readable string with the time statistics for this
        keeper. The operations are sorted by decreasing average time.

        Example result, if keeper is enabled:

        ::

            Time statistics (times in seconds):
            Count  Average  Minimum  Maximum  Operation name
                1  0.024    0.024    0.024    get /api/cpcs
                1  0.009    0.009    0.009    get /api/version
        """
        ret = "Time statistics (times in seconds):\n"
        if self.enabled:
            ret += "Count  Average  Minimum  Maximum  Operation name\n"
            snapshot_by_avg = sorted(self.snapshot(),
                                     key=lambda item: item[1].avg_time,
                                     reverse=True)
            for name, stats in snapshot_by_avg:
                ret += "{:5d}  {:7.3f}  {:7.3f}  {:7.3f}  {}\n".format(
                    stats.count, stats.avg_time, stats.min_time,
                    stats.max_time, name)
        else:
            ret += "Disabled.\n"
        return ret.strip()



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Exceptions that can be raised by the client.
"""

import re


__all__ = ['Error', 'ConnectionError', 'ConnectTimeout', 'ReadTimeout',
           'RetriesExceeded', 'AuthError', 'ParseError', 'VersionError',
           'HTTPError', 'OperationTimeout', 'StatusTimeout',
           'NoUniqueMatch', 'NotFound']


class Error(Exception):
    """
    Abstract base class for exceptions specific to this package.

    Derived from :exc:`~py:exceptions.Exception`.
    """

    def __init__(self, *args):
        """
        Parameters:

          *args:
            A list of input arguments for the exception object.

            The derived classes define more specific parameters.

            These input arguments will be available as tuple items in the
            ``args`` instance variable of the exception object.
        """
        super(Error, self).__init__(*args)


class ConnectionError(Error):
    """
    This exception indicates a problem with the connection to the HMC, below
    the HTTP level. HTTP errors are indicated via :exc:`~zhmcclient.HTTPError`.

    A retry by the user code is not likely to be successful, unless connect or
    read retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, details=None):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The original exception describing details about the error.
        """
        super(ConnectionError, self).__init__(msg)
        self._details = details

    @property
    def details(self):
        """
        The original exception caught by this package, providing more
        information about the problem, if there was such an original exception.
        `None`, otherwise.

        This may be one of the following exceptions:

        * Any exception derived from
          :exc:`requests.exceptions.RequestException`.
        """
        return self._details


class ConnectTimeout(ConnectionError):
    """
    This exception indicates that a connection to the HMC timed out after
    exhausting the connect retries (see
    :attr:`zhmcclient.RetryTimeoutConfig.connect_retries`).

    Further retrying by the user code is not likely to be successful, unless
    connect retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.ConnectionError`.
    """


class ReadTimeout(ConnectionError):
    """
    This exception indicates that reading an HTTP response from the HMC timed
    out after exhausting the read retries (see
    :attr:`zhmcclient.RetryTimeoutConfig.read_retries`).

    Further retrying by the user code is not likely to be successful, unless
    read retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.ConnectionError`.
    """


class RetriesExceeded(ConnectionError):
    """
    This exception indicates that the maximum number of retries for connecting
    to the HMC, sending HTTP requests or reading HTTP responses was exceeded,
    for reasons other than connect timeouts (see
    :exc:`~zhmcclient.ConnectTimeout`) or read timeouts (see
    :exc:`~zhmcclient.ReadTimeout`).

    Further retrying by the user code is not likely to be successful, unless
    connect or read retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.ConnectionError`.
    """


class AuthError(Error):
    """
    This exception indicates an authentication error with the HMC, either at
    the TLS/SSL handshake level (e.g. with CA certificates), or at the HTTP
    level.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, details=None):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The original exception describing details about the error.
        """
        super(AuthError, self).__init__(msg)
        self._details = details

    @property
    def details(self):
        """
        The original exception describing details about the error, if there
        was such an original exception. `None`, otherwise.

        This may be one of the following exceptions:

        * :exc:`~zhmcclient.HTTPError`
        """
        return self._details


class ParseError(Error):
    """
    This exception indicates a parsing error while processing the JSON payload
    in a response from the HMC.

    Derived from :exc:`~zhmcclient.Error`.

    The error location within the payload is automatically determined by
    parsing the error message for the pattern::

      ``': line 1 column 2 '``
    """

    def __init__(self, msg):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

            This should be the message of the `ValueError` exception raised
            by methods of the :class:`py:json.JSONDecoder` class.
        """
        super(ParseError, self).__init__(msg)
        m = re.search(r': line ([0-9]+) column ([0-9]+) ', msg)
        if m:
            self._line = int(m.group(1))
            self._column = int(m.group(2))
        else:
            self._line = None
            self._column = None

    @property
    def line(self):
        """
        The 1-based line number of the error location within the JSON payload,
        as an integer.

        `None` indicates that the error location is not available.
        """
        return self._line

    @property
    def column(self):
        """
        The 1-based column number of the error location within the JSON
        payload, as an integer.

        `None` indicates that the error location is not available.
        """
        return self._column


class VersionError(Error):
    """
    This exception indicates that a function of the client requires a minimum
    HMC API version which is not supported by the HMC.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, min_api_version, api_version):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          min_api_version (:term:`HMC API version`):
            The minimum HMC API version required to perform the function that
            raised this exception.

          api_version (:term:`HMC API version`):
            The actual HMC API version supported by the HMC.
        """
        super(VersionError, self).__init__(msg)
        self._min_api_version = min_api_version
        self._api_version = api_version

    @property
    def min_api_version(self):
        """
        :term:`HMC API version`: The minimum HMC API version required to
        perform the function that raised this exception.
        """
        return self._min_api_version

    @property
    def api_version(self):
        """
        :term:`HMC API version`: The actual HMC API version supported by the
        HMC.
        """
        return self._api_version


class HTTPError(Error):
    """
    This exception indicates that the HMC returned an HTTP response with a bad
    HTTP status code.

    Derived from :exc:`~zhmcclient.Error`.
    """

    html_error_reason = 999  # Used when the HMC returns HTML content

    def __init__(self, body):
        """
        Parameters:

          body (:term:`json object`):
            Body of the HTTP error response.
        """
        super(HTTPError, self).__init__(body)
        self._body = body

    @property
    def http_status(self):
        """
        :term:`integer`: Numeric HTTP status code (e.g. 500).

        See :term:`RFC2616` for a list of HTTP status codes and reason phrases.
        """
        return self._body.get('http-status', None)

    @property
    def reason(self):
        """
        :term:`integer`: Numeric HMC reason code.

        HMC reason codes provide more details as to the nature of the error
        than is provided by the HTTP status code. This HMC reason code is
        treated as a sub-code of the HTTP status code and thus must be used in
        conjunction with the HTTP status code to determine the error condition.

        Standard HMC reason codes that apply across the entire API are
        described in section 'Common request validation reason codes' in the
        :term:`HMC API` book. Additional operation-specific reason codes may
        also be documented in the description of specific API operations in the
        :term:`HMC API` book.

        The articial reason code 999 is used when the response from the HMC
        contains an HTML-formatted error message.
        """
        return self._body.get('reason', None)

    @property
    def message(self):
        """
        :term:`string`: Message describing the error.

        This message is not currently localized.
        """
        return self._body.get('message', None)

    @property
    def request_method(self):
        """
        :term:`string`: The HTTP method (DELETE, GET, POST, PUT) that caused
        this error response.
        """
        return self._body.get('request-method', None)

    @property
    def request_uri(self):
        """
        :term:`string`: The URI that caused this error response.
        """
        return self._body.get('request-uri', None)

    @property
    def request_query_parms(self):
        """
        List of query-parm-info objects: URI query parameters specified on the
        request.

        Each query-parm-info object identifies a single query parameter by its
        name and includes its value(s).

        An empty list, if the request did not specify any query parameters.
        """
        return self._body.get('request-query-parms', None)

    @property
    def request_headers(self):
        """
        header-info object: HTTP headers specified on the request.

        An empty list, if the request did not specify any HTTP headers.
        """
        return self._body.get('request-headers', None)

    @property
    def request_authenticated_as(self):
        """
        :term:`string`: Name of the HMC user associated with the API session
        under which the request was issued.

        `None`, if the request was issued without an established session or
        there is no HMC user bound to the session.
        """
        return self._body.get('request-authenticated-as', None)

    @property
    def request_body(self):
        """
        The request body, in the form of a JSON document. Note that, since it
        is in the form of a JSON document, this may not be exactly what was
        submitted by the API client program, but it is semantically equivalent.

        If the request body could not be parsed or some other error prevented
        the creation of a JSON document from the request body, this property
        is `None` and the request body is instead available in the
        :attr:`~zhmcclient.HTTPError.request_body_as_string` property.
        """
        return self._body.get('request-body', None)

    @property
    def request_body_as_string(self):
        """
        :term:`string`: The complete request body, or some portion of the
        request body, exactly as it was submitted by the API client program, if
        the :attr:`~zhmcclient.HTTPError.request_body` property is `None`.
        Otherwise, `None`.

        The :attr:`~zhmcclient.HTTPError.request_body_as_string_partial`
        property indicates whether the complete request body is provided in
        this property.
        """
        return self._body.get('request-body-as-string', None)

    @property
    def request_body_as_string_partial(self):
        """
        :class:`py:bool`: Indicates whether the
        :attr:`~zhmcclient.HTTPError.request_body_as_string` property contains
        only part of the request body (`True`) or the entire request body
        (`False`). `None`, if the
        :attr:`~zhmcclient.HTTPError.request_body_as_string` property is
        `None`.
        """
        return self._body.get('request-body-as-string-partial', None)

    @property
    def stack(self):
        """
        :term:`string`: Internal HMC diagnostic information for the error.

        This field is supplied only on selected 5xx HTTP status codes.
        `None`, if not supplied.
        """
        return self._body.get('stack', None)

    @property
    def error_details(self):
        """
        :term:`string`: A nested object that provides additional
        operation-specific error information. This field is provided by
        selected operations, and the format of the nested object is as
        described by that operation.
        """
        return self._body.get('error-details', None)

    def __str__(self):
        return "{},{}: {} [{} {}]".\
               format(self.http_status, self.reason, self.message,
                      self.request_method, self.request_uri)

    def __repr__(self):
        return "HTTPError(http_status={}, reason={}, message={}, "\
               "request_method={}, request_uri={}, ...)".\
               format(self.http_status, self.reason, self.message,
                      self.request_method, self.request_uri)


class OperationTimeout(Error):
    """
    This exception indicates that the waiting for completion of an asynchronous
    HMC operation has timed out.

    Derived from :exc:`~zhmcclient.Error`.
    """
    pass


class StatusTimeout(Error):
    """
    This exception indicates that the waiting for reaching a desired resource
    status has timed out.

    Derived from :exc:`~zhmcclient.Error`.
    """
    pass


class NoUniqueMatch(Error):
    """
    This exception indicates that a find function has found more than one item.

    Derived from :exc:`~zhmcclient.Error`.
    """
    pass


class NotFound(Error):
    """
    This exception indicates that a find function did not find an item.

    Derived from :exc:`~zhmcclient.Error`.
    """
    pass



# -*- coding: utf-8 -*-
#
# Configuration file for Sphinx builds for the zhmcclient project.
#
# Originally created by sphinx-quickstart, but then manually maintained.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import re

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.3'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',   # disabed, raises anexception
    'sphinx.ext.ifconfig',
    'sphinx_git',            # requires 'sphinx-git' Python package
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8'

# The master toctree document.
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    master_doc = 'index'
else:
    master_doc = 'docs/index'

# General information about the project.
project = u'zhmcclient'
copyright = u'IBM'
author = u'z Systems KVM OpenStack Team'

# The short description of the package.
_short_description = u'Client library for z Systems Hardware Management Console Web Services API'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

def package_version(filename, varname):
    """Return package version string by reading `filename` and retrieving its
       module-global variable `varnam`."""
    _locals = {}
    with open(filename) as fp:
        exec(fp.read(), None, _locals)
    return _locals[varname]

# The short X.Y version.
# Note: We use the full version in both cases (e.g. 'M.N.U' or 'M.N.U.dev0').
version = package_version('../zhmcclient/_version.py', '__version__')

# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["tests", ".tox", ".git", "attic", "dist",
                    "build_doc", "zhmcclient.egg-info", ".eggs"]

# The reST default role (used for this markup: `text`) to use for all
# documents. None means it is rendered in italic, without a link.
default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for Napoleon extension ---------------------------------------

# Enable support for Google style docstrings. Defaults to True.
napoleon_google_docstring = True

# Enable support for NumPy style docstrings. Defaults to True.
napoleon_numpy_docstring = False

# Include private members (like _membername). False to fall back to Sphinx’s
# default behavior. Defaults to False.
napoleon_include_private_with_doc = False

# Include special members (like __membername__). False to fall back to Sphinx’s
# default behavior. Defaults to True.
napoleon_include_special_with_doc = True

# Use the .. admonition:: directive for the Example and Examples sections,
# instead of the .. rubric:: directive. Defaults to False.
napoleon_use_admonition_for_examples = False

# Use the .. admonition:: directive for Notes sections, instead of the
# .. rubric:: directive. Defaults to False.
napoleon_use_admonition_for_notes = False

# Use the .. admonition:: directive for References sections, instead of the
# .. rubric:: directive. Defaults to False.
napoleon_use_admonition_for_references = False

# Use the :ivar: role for instance variables, instead of the .. attribute::
# directive. Defaults to False.
napoleon_use_ivar = True

# Use a :param: role for each function parameter, instead of a single
# :parameters: role for all the parameters. Defaults to True.
napoleon_use_param = True

# Use the :rtype: role for the return type, instead of inlining it with the
# description. Defaults to True.
napoleon_use_rtype = True


# -- Options for viewcode extension ---------------------------------------

# Follow alias objects that are imported from another module such as functions,
# classes and attributes. As side effects, this option ... ???
# If false, ... ???.
# The default is True.
viewcode_import = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
# See http://www.sphinx-doc.org/en/stable/theming.html for built-in themes.
html_theme = "classic"

# Theme options are theme-specific and customize the look and feel of a theme
# further.
# See http://www.sphinx-doc.org/en/stable/theming.html for the options
# available for built-in themes.
html_theme_options = {
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If not defined, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = 'ld'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
html_extra_path = ['_extra']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = project+'_doc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
'papersize': 'a4paper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, project+'.tex', _short_description, author, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, project, _short_description, [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, project, _short_description,
     author, project, _short_description,
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# -- Options for autodoc extension ----------------------------------------
# For documentation, see
# http://www.sphinx-doc.org/en/stable/ext/autodoc.html

# Selects what content will be inserted into a class description.
# The possible values are:
#   "class" - Only the class’ docstring is inserted. This is the default.
#   "both"  - Both the class’ and the __init__ method’s docstring are
#             concatenated and inserted.
#   "init"  - Only the __init__ method’s docstring is inserted.
# In all cases, the __init__ method is still independently rendered as a
# special method, e.g. when the :special-members: option is set.
autoclass_content = "both"

# Selects if automatically documented members are sorted alphabetically
# (value 'alphabetical'), by member type (value 'groupwise') or by source
# order (value 'bysource'). The default is alphabetical.
autodoc_member_order = "bysource"

# This value is a list of autodoc directive flags that should be automatically
# applied to all autodoc directives. The supported flags are:
# 'members', 'undoc-members', 'private-members', 'special-members',
# 'inherited-members' and 'show-inheritance'.
# If you set one of these flags in this config value, you can use a negated
# form, 'no-flag', in an autodoc directive, to disable it once.
autodoc_default_flags = []

# Functions imported from C modules cannot be introspected, and therefore the
# signature for such functions cannot be automatically determined. However, it
# is an often-used convention to put the signature into the first line of the
# function’s docstring.
# If this boolean value is set to True (which is the default), autodoc will
# look at the first line of the docstring for functions and methods, and if it
# looks like a signature, use the line as the signature and remove it from the
# docstring content.
autodoc_docstring_signature = True

# This value contains a list of modules to be mocked up. This is useful when
# some external dependencies are not met at build time and break the building
# process.
autodoc_mock_imports = []


# -- Options for intersphinx extension ------------------------------------
# For documentation, see
# http://www.sphinx-doc.org/en/stable/ext/intersphinx.html

# Defines the prefixes for intersphinx links, and the targets they resolve to.
# Example RST source for 'py2' prefix:
#     :func:`py2:platform.dist`
#
# Note: The URLs apparently cannot be the same for two different IDs; otherwise
#       the links for one of them are not being created. A small difference
#       such as adding a trailing backslash is already sufficient to work
#       around the problem.
#
# Note: This mapping does not control how links to datatypes of function
#       parameters are generated.
# TODO: Find out how the targeted Python version for auto-generated links
#       to datatypes of function parameters can be controlled.
#
intersphinx_mapping = {
  'py': ('https://docs.python.org/2/', None), # agnostic to Python version
  'py2': ('https://docs.python.org/2', None), # specific to Python 2
  'py3': ('https://docs.python.org/3', None), # specific to Python 3
}

intersphinx_cache_limit = 5

# -- Options for extlinks extension ---------------------------------------
# For documentation, see
# http://www.sphinx-doc.org/en/stable/ext/extlinks.html
#
# Defines aliases for external links that can be used as role names.
#
# This config value must be a dictionary of external sites, mapping unique
# short alias names to a base URL and a prefix:
# * key: alias-name
# * value: tuple of (base-url, prefix)
#
# Example for the config value:
#
#   extlinks = {
#     'issue': ('https://github.com/sphinx-doc/sphinx/issues/%s', 'Issue ')
#   }
#
# The alias-name can be used as a role in links. In the example, alias name
# 'issue' is used in RST as follows:
#   :issue:`123`.
# This then translates into a link:
#   https://github.com/sphinx-doc/sphinx/issues/123
# where the %s in the base-url was replaced with the value between back quotes.
#
# The prefix plays a role only for the link caption:
# * If the prefix is None, the link caption is the full URL.
# * If the prefix is the empty string, the link caption is the partial URL
#   given in the role content ("123" in this case.)
# * If the prefix is a non-empty string, the link caption is the partial URL,
#   prepended by the prefix. In the above example, the link caption would be
#   "Issue 123".
#
# You can also use the usual "explicit title" syntax supported by other roles
# that generate links to set the caption. In this case, the prefix is not
# relevant.
# For example, this RST:
#   :issue:`this issue <123>`
# results in the link caption "this issue".

extlinks = {
  'nbview': ('http://nbviewer.jupyter.org/github/zhmcclient/python-zhmcclient/blob/master/docs/notebooks/%s', ''),
  'nbdown': ('https://github.com/zhmcclient/python-zhmcclient/raw/master/docs/notebooks/%s', '')
}




# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An :term:`Adapter` is a physical adapter card (e.g. OSA-Express adapter,
Crypto adapter) or a logical adapter (e.g. HiperSockets switch).

Adapter resources are contained in :term:`CPC` resources.

Adapters only exist in CPCs that are in DPM mode.

There are four types of Adapters:

1. Network Adapters:
   Network adapters enable communication through different networking
   transport protocols. These network adapters are OSA-Express,
   HiperSockets and RoCE-Express.
   DPM automatically discovers OSA-Express and RoCE-Express adapters
   because they are physical cards that are installed on the CPC.
   In contrast, HiperSockets are logical adapters and must be
   created and configured by an administrator using the 'Create Hipersocket'
   operation (see create_hipersocket()).
   Network Interface Cards (NICs) provide a partition with access to networks.
   Each NIC represents a unique connection between the partition
   and a specific network adapter.

2. Storage Adapters:
   Fibre Channel connections provide high-speed connections between CPCs
   and storage devices.
   DPM automatically discovers any storage adapters installed on the CPC.
   Host bus adapters (HBAs) provide a partition with access to external
   storage area networks (SANs) and devices that are connected to a CPC.
   Each HBA represents a unique connection between the partition
   and a specific storage adapter.

3. Accelerator Adapters:
   Accelerator adapters provide specialized functions to
   improve performance or use of computer resource like the IBM System z
   Enterprise Data Compression (zEDC) feature.
   DPM automatically discovers accelerators that are installed on the CPC.
   An accelerator virtual function provides a partition with access
   to zEDC features that are installed on a CPC.
   Each virtual function represents a unique connection between
   the partition and a physical feature card.

4. Crypto Adapters:
   Crypto adapters provide cryptographic processing functions.
   DPM automatically discovers cryptographic features that are installed
   on the CPC.
"""

from __future__ import absolute_import

from ._manager import BaseManager
from ._resource import BaseResource
from ._port import PortManager
from ._logging import get_logger, logged_api_call

__all__ = ['AdapterManager', 'Adapter']

LOG = get_logger(__name__)


class AdapterManager(BaseManager):
    """
    Manager providing access to the :term:`Adapters <Adapter>` in a particular
    :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object (in DPM mode):

    * :attr:`~zhmcclient.Cpc.adapters`
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
            'adapter-id',
            'adapter-family',
            'type',
            'status',
        ]

        super(AdapterManager, self).__init__(
            resource_class=Adapter,
            parent=cpc,
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the Adapters in this CPC.

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any Adapter to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Adapter` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms, client_filters = self._divide_filter_args(filter_args)

        resources_name = 'adapters'
        uri = '{}/{}{}'.format(self.cpc.uri, resources_name, query_parms)

        resource_obj_list = []
        result = self.session.get(uri)
        if result:
            props_list = result[resources_name]
            for props in props_list:

                resource_obj = self.resource_class(
                    manager=self,
                    uri=props[self._uri_prop],
                    name=props.get(self._name_prop, None),
                    properties=props)

                if self._matches_filters(resource_obj, client_filters):
                    resource_obj_list.append(resource_obj)
                    if full_properties:
                        resource_obj.pull_full_properties()

        return resource_obj_list

    @logged_api_call
    def create_hipersocket(self, properties):
        """
        Create and configure a HiperSockets Adapter in this CPC.

        Authorization requirements:

        * Object-access permission to the scoping CPC.
        * Task permission to the "Create HiperSockets Adapter" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Hipersocket' in the :term:`HMC API` book.

        Returns:

          :class:`~zhmcclient.Adapter`:
            The resource object for the new HiperSockets Adapter.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.cpc.uri + '/adapters', body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = properties.copy()
        props.update(result)
        return Adapter(self, props['object-uri'], None, props)


class Adapter(BaseResource):
    """
    Representation of an :term:`Adapter`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of an Adapter, see section 'Data model' in section
    'Adapter object' in the :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.AdapterManager`).
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.AdapterManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        if not isinstance(manager, AdapterManager):
            raise AssertionError("Adapter init: Expected manager type %s, "
                                 "got %s" %
                                 (AdapterManager, type(manager)))
        super(Adapter, self).__init__(manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._ports = None

    @property
    def ports(self):
        """
        :class:`~zhmcclient.PortManager`: Access to the :term:`Ports <Port>` of
        this Adapter.
        """
        # We do here some lazy loading.
        if not self._ports:
            self._ports = PortManager(self)
        return self._ports

    @logged_api_call
    def delete(self):
        """
        Delete this Adapter.

        The Adapter must be a HiperSockets Adapter.

        Authorization requirements:

        * Object-access permission to the HiperSockets Adapter to be deleted.
        * Task permission to the "Delete HiperSockets Adapter" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        self.manager.session.delete(self.uri)

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Adapter.

        Authorization requirements:

        * Object-access permission to the Adapter.
        * Task permission for the "Adapter Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Adapter object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        self.manager.session.post(self.uri, body=properties)



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Port` is a physical connector port (jack) of an :term:`Adapter`.

Port resources are contained in Adapter resources.

Ports only exist in :term:`CPCs <CPC>` that are in DPM mode.
"""

from __future__ import absolute_import

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import get_logger, logged_api_call

__all__ = ['PortManager', 'Port']

LOG = get_logger(__name__)


class PortManager(BaseManager):
    """
    Manager providing access to the :term:`Ports <Port>` of a particular
    :term:`Adapter`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible as properties in higher level resources (in this case, the
    :class:`~zhmcclient.Adapter` object).
    """

    def __init__(self, adapter):
        # This function should not go into the docs.
        # Parameters:
        #   adapter (:class:`~zhmcclient.Adapter`):
        #     Adapter defining the scope for this manager.

        super(PortManager, self).__init__(
            resource_class=Port,
            parent=adapter,
            uri_prop='element-uri',
            name_prop='name',
            query_props=[])

    @property
    def adapter(self):
        """
        :class:`~zhmcclient.Adapter`: :term:`Adapter` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the Ports of this Adapter.

        Authorization requirements:

        * Object-access permission to this Adapter.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Port` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        resource_obj_list = []
        storage_family = ('ficon')
        network_family = ('osa', 'roce', 'hipersockets')
        if self.adapter.get_property('adapter-family') in storage_family:
            uris = self.adapter.get_property('storage-port-uris')
        elif self.adapter.get_property('adapter-family') in network_family:
            uris = self.adapter.get_property('network-port-uris')
        else:
            return resource_obj_list
        if uris:
            for uri in uris:

                resource_obj = self.resource_class(
                    manager=self,
                    uri=uri,
                    name=None,
                    properties=None)

                if self._matches_filters(resource_obj, filter_args):
                    resource_obj_list.append(resource_obj)
                    if full_properties:
                        resource_obj.pull_full_properties()
        return resource_obj_list


class Port(BaseResource):
    """
    Representation of a :term:`Port`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of a Port, see section
    'Data model - Port Element Object' in section 'Adapter object' in the
    :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.PortManager`).
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.PortManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        if not isinstance(manager, PortManager):
            raise AssertionError("Port init: Expected manager type %s, "
                                 "got %s" %
                                 (PortManager, type(manager)))
        super(Port, self).__init__(manager, uri, name, properties)

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Port.

        Authorization requirements:

        * Object-access permission to the Adapter of this Port.
        * Task permission to the "Adapter Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model - Port Element Object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        self.manager.session.post(self._uri, body=properties)



# Copyright 2016 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`CPC` (Central Processor Complex) is a physical z Systems or LinuxONE
computer.

A particular HMC can manage multiple CPCs.

The HMC can manage a range of old and new CPC generations. Some older CPC
generations are not capable of supporting the HMC Web Services API; these older
CPCs can be managed using the GUI of the HMC, but not through its Web Services
API. Therefore, such older CPCs will not be exposed at the HMC Web Services
API, and thus will not show up in the API of this Python package.

TODO: List earliest CPC generation that supports the HMC Web Services API.

A CPC can be in any of the following three modes:

- DPM mode: Dynamic Partition Manager is enabled for the CPC.
- Classic mode: The CPC does not have Dynamic Partition Manager enabled,
  and is not member of an ensemble.
- Ensemble mode: The CPC is member of an ensemble. This Python client
  does not support the functionality that is specific to ensemble mode.

The functionality supported at the HMC API and thus also for users of this
Python client, depends on the mode in which the CPC currently is. If a
particular functionality is available only in a specific mode, that is
indicated in the description of the functionality.
"""

from __future__ import absolute_import

from ._manager import BaseManager
from ._resource import BaseResource
from ._lpar import LparManager
from ._partition import PartitionManager
from ._activation_profile import ActivationProfileManager
from ._adapter import AdapterManager
from ._virtual_switch import VirtualSwitchManager
from ._logging import get_logger, logged_api_call
from ._exceptions import HTTPError

__all__ = ['CpcManager', 'Cpc']

LOG = get_logger(__name__)


class CpcManager(BaseManager):
    """
    Manager providing access to the :term:`CPCs <CPC>` exposed by the HMC this
    client is connected to.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Client` object:

    * :attr:`~zhmcclient.Client.cpcs`
    """

    def __init__(self, client):
        # This function should not go into the docs.
        # Parameters:
        #   client (:class:`~zhmcclient.Client`):
        #      Client object for the HMC to be used.

        # Resource properties that are supported as filter query parameters
        # (for server-side filtering).
        query_props = [
            'name',
        ]

        super(CpcManager, self).__init__(
            resource_class=Cpc,
            parent=None,
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

        self._session = client.session

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the CPCs exposed by the HMC this client is connected to.

        Authorization requirements:

        * Object-access permission to any CPC to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Cpc` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms, client_filters = self._divide_filter_args(filter_args)

        resources_name = 'cpcs'
        uri = '/api/{}{}'.format(resources_name, query_parms)

        resource_obj_list = []
        result = self.session.get(uri)
        if result:
            props_list = result[resources_name]
            for props in props_list:

                resource_obj = self.resource_class(
                    manager=self,
                    uri=props[self._uri_prop],
                    name=props.get(self._name_prop, None),
                    properties=props)

                if self._matches_filters(resource_obj, client_filters):
                    resource_obj_list.append(resource_obj)
                    if full_properties:
                        resource_obj.pull_full_properties()

        return resource_obj_list


class Cpc(BaseResource):
    """
    Representation of a :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.CpcManager`).
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.CpcManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        if not isinstance(manager, CpcManager):
            raise AssertionError("Cpc init: Expected manager type %s, got %s" %
                                 (CpcManager, type(manager)))
        super(Cpc, self).__init__(manager, uri, name, properties)

        # The manager objects for child resources (with lazy initialization):
        self._lpars = None
        self._partitions = None
        self._adapters = None
        self._virtual_switches = None
        self._reset_activation_profiles = None
        self._image_activation_profiles = None
        self._load_activation_profiles = None

    @property
    def lpars(self):
        """
        :class:`~zhmcclient.LparManager`: Access to the :term:`LPARs <LPAR>` in
        this CPC.
        """
        # We do here some lazy loading.
        if not self._lpars:
            self._lpars = LparManager(self)
        return self._lpars

    @property
    def partitions(self):
        """
        :class:`~zhmcclient.PartitionManager`: Access to the
        :term:`Partitions <Partition>` in this CPC.
        """
        # We do here some lazy loading.
        if not self._partitions:
            self._partitions = PartitionManager(self)
        return self._partitions

    @property
    def adapters(self):
        """
        :class:`~zhmcclient.AdapterManager`: Access to the
        :term:`Adapters <Adapter>` in this CPC.
        """
        # We do here some lazy loading.
        if not self._adapters:
            self._adapters = AdapterManager(self)
        return self._adapters

    @property
    def virtual_switches(self):
        """
        :class:`~zhmcclient.VirtualSwitchManager`: Access to the
        :term:`Virtual Switches <Virtual Switch>` in this CPC.
        """
        # We do here some lazy loading.
        if not self._virtual_switches:
            self._virtual_switches = VirtualSwitchManager(self)
        return self._virtual_switches

    @property
    def vswitches(self):
        """
        Deprecated: Use :attr:`~zhmcclient.Cpc.virtual_switches` instead.
        """
        # TODO: Issue a deprecation warning
        return self.virtual_switches

    @property
    def reset_activation_profiles(self):
        """
        :class:`~zhmcclient.ActivationProfileManager`: Access to the
        :term:`Reset Activation Profiles <Reset Activation Profile>` in this
        CPC.
        """
        # We do here some lazy loading.
        if not self._reset_activation_profiles:
            self._reset_activation_profiles = \
                ActivationProfileManager(self, profile_type='reset')
        return self._reset_activation_profiles

    @property
    def image_activation_profiles(self):
        """
        :class:`~zhmcclient.ActivationProfileManager`: Access to the
        :term:`Image Activation Profiles <Image Activation Profile>` in this
        CPC.
        """
        # We do here some lazy loading.
        if not self._image_activation_profiles:
            self._image_activation_profiles = \
                ActivationProfileManager(self, profile_type='image')
        return self._image_activation_profiles

    @property
    def load_activation_profiles(self):
        """
        :class:`~zhmcclient.ActivationProfileManager`: Access to the
        :term:`Load Activation Profiles <load Activation Profile>` in this
        CPC.
        """
        # We do here some lazy loading.
        if not self._load_activation_profiles:
            self._load_activation_profiles = \
                ActivationProfileManager(self, profile_type='load')
        return self._load_activation_profiles

    @property
    @logged_api_call
    def dpm_enabled(self):
        """
        bool: Indicates whether this CPC is currently in DPM mode
        (Dynamic Partition Manager mode).

        If the CPC is not currently in DPM mode, or if the CPC does not
        support DPM mode (i.e. before z13), `False` is returned.

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to all Partitions of this CPC, when the
          CPC is in DPM mode.
        * Object-access permission to all LPARs of this CPC, when the
          CPC is in classic (or ensemble) mode.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # This is an optimized implementation that is faster than the
        # "Get CPC Properties" operation (which would return the information
        # directly in its "dpm-enabled" property).
        try:
            # We perform a test operation that only succeeds when the CPC is in
            # DPM mode:
            self.partitions.list()
            return True
        except HTTPError as exc:
            if exc.http_status == 404 and exc.reason == 1:
                # Description of this error: The request URI does not designate
                # an existing resource of the expected type, or designates a
                # resource for which the API user does not have object-access
                # permission.
                # Because this error response does not distinguish between
                # these two cases, we need to perform the opposite test
                # operation as well:
                self.lpars.list()
                return False
            else:
                raise

    @logged_api_call
    def start(self, wait_for_completion=True, operation_timeout=None):
        """
        Start this CPC, using the HMC operation "Start CPC".

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission for the "Start (start a single DPM system)" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        result = self.manager.session.post(
            self.uri + '/operations/start',
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def stop(self, wait_for_completion=True, operation_timeout=None):
        """
        Stop this CPC, using the HMC operation "Stop CPC".

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission for the "Stop (stop a single DPM system)" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        result = self.manager.session.post(
            self.uri + '/operations/stop',
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def import_profiles(self, profile_area, wait_for_completion=True,
                        operation_timeout=None):
        """
        Import activation profiles and/or system activity profiles for this CPC
        from the SE hard drive into the CPC using the HMC operation
        "Import Profiles".

        This operation is not permitted when the CPC is in DPM mode.

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission for the "CIM Actions ExportSettingsData" task.

        Parameters:

          profile_area (int):
            The numbered hard drive area (1-4) from which the profiles are
            imported.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        body = {'profile-area': profile_area}
        result = self.manager.session.post(
            self.uri + '/operations/import-profiles',
            body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def export_profiles(self, profile_area, wait_for_completion=True,
                        operation_timeout=None):
        """
        Export activation profiles and/or system activity profiles from this
        CPC to the SE hard drive using the HMC operation "Export Profiles".

        This operation is not permitted when the CPC is in DPM mode.

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission for the "CIM Actions ExportSettingsData" task.

        Parameters:

          profile_area (int):
             The numbered hard drive area (1-4) to which the profiles are
             exported. Any existing data is overwritten.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        body = {'profile-area': profile_area}
        result = self.manager.session.post(
            self.uri + '/operations/export-profiles',
            body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def get_wwpns(self, partitions):
        """
        Return the WWPNs of the host ports (of the :term:`HBAs <HBA>`) of the
        specified :term:`Partitions <Partition>` of this CPC.

        This method performs the HMC operation "Export WWPN List".

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to the Partitions designated by the
          "partitions" parameter.
        * Task permission for the "Export WWPNs" task.

        Parameters:

          partitions (:term:`iterable` of :class:`~zhmcclient.Partition`):
            :term:`Partitions <Partition>` to be used.

        Returns:

          A list of items for each WWPN, where each item is a dict with the
          following keys:

          * 'partition-name' (string): Name of the :term:`Partition`.
          * 'adapter-id' (string): ID of the :term:`FCP Adapter`.
          * 'device-number' (string): Virtual device number of the :term:`HBA`.
          * 'wwpn' (string): WWPN of the HBA.

        Raises:

          :exc:`~zhmcclient.HTTPError`: See the HTTP status and reason codes of
            operation "Export WWPN List" in the :term:`HMC API` book.
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'partitions': [p.uri for p in partitions]}
        result = self.manager.session.post(self._uri + '/operations/'
                                           'export-port-names-list', body=body)
        # Parse the returned comma-separated string for each WWPN into a dict:
        wwpn_list = []
        dict_keys = ('partition-name', 'adapter-id', 'device-number', 'wwpn')
        for wwpn_item in result['wwpn-list']:
            dict_values = wwpn_item.split(',')
            wwpn_list.append(dict(zip(dict_keys, dict_values)))
        return wwpn_list
